"""Check command."""
import os
from typing import List, Tuple, Optional

import arrow
import click
import docker
from docker import DockerClient
from docker.errors import ImageNotFound
from loguru import logger

from megalus.main import Megalus

client = docker.from_env()


def get_services_to_check(meg: Megalus, service: str, service_list: List[str], ignore_list: List[str], tree: List[str],
                          compose_data: dict) -> Optional[Tuple[List, List]]:
    """Get services to check.

    Recursively get all services to check and ignore.

    :param meg: click context object
    :param service: service to inspect
    :param service_list: services to check list
    :param ignore_list: services to ignore list
    :param tree: current service dependencies list
    :param compose_data: docker-compose data
    :return: Tuple
    """
    if service in service_list or service in ignore_list:
        return  # type: ignore

    service_list.append(service) if compose_data.get('build', None) else ignore_list.append(  # type: ignore
        service)
    if compose_data.get('depends_on', None):
        tree = compose_data['depends_on']

    if tree:
        for key in tree:
            key_data = meg.find_service(key)
            get_services_to_check(meg, key, service_list, ignore_list, tree, key_data['compose_data'])
    return service_list, ignore_list


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.pass_obj
def check(meg: Megalus, services) -> None:
    """Check services.

    This command will check the selected services for:

    * Need build because does not have image
    * Need build because his image is outdated

    To find out if the image needs updating,
    add the list of files in the megalus.yml file,
    whose update date should be compared to the build date of the image.

    Example: if you add the 'Dockerfile', 'requirements.txt' and 'Pipfile'
    inside the 'check_for_build: files:' section in megalus.yml, this command will compare,
    for each file, his last modification date against docker image build date
    for the service.

    :param meg: click context object
    :param services: services list to be inspected
    :return: None
    """
    def get_docker_image(compose_data: dict) -> Optional[DockerClient]:  # type: ignore
        """Get docker image data.

        :param compose_data: compose parsed data.
        :return: DockerClient image instance or None
        """
        if compose_data.get('image', None) and compose_data.get('build', None):
            try:
                return client.images.get(compose_data['image'])
            except ImageNotFound:
                return None

    def has_old_image(ctx: Megalus, service: str) -> bool:
        """Check if image is outdated.

        :param ctx: Megalus instance
        :param service: docker service selected
        :return: bool
        """
        def get_date_from_file(file: str) -> arrow:
            """Get date from file.

            :param file: file full path
            :return: Arrow instance
            """
            date = arrow.get(os.path.getmtime(os.path.join(data['working_dir'], file))).to('local')
            logger.debug("Last update for file {} is {}".format(file, date))
            return date

        data = ctx.find_service(service)
        image = get_docker_image(data['compose_data'])
        if not image:
            return False
        else:
            image_date_created = arrow.get(image.attrs['Created']).to('local')
            global_files_to_watch = meg.config_data['defaults'].get('check_for_build', [])
            list_dates = [
                get_date_from_file(file)
                for file in global_files_to_watch
                if os.path.isfile(os.path.join(data['working_dir'], file))
                # FIXME: Acertar o working_dir + build.context para achar o path dos arquivos
            ]
            if list_dates and image_date_created < max(list_dates):
                return True
            return False

    service_list = []  # type: List[str]
    ignore_list = []  # type: List[str]
    for service in services:
        logger.info("Checking {}...".format(service))
        service_data = meg.find_service(service)
        service_list, ignore_list = get_services_to_check(  # type: ignore
            meg, service, service_list, ignore_list, [],
            service_data['compose_data']
        )

    services_without_images = [
        service
        for service in service_list
        if get_docker_image(meg.find_service(service))
    ]

    if services_without_images:
        logger.warning("Services without images: {}".format(", ".join(services_without_images)))

    services_with_old_images = [
        service
        for service in service_list
        if service not in services_without_images and has_old_image(meg, service)
    ]

    if services_with_old_images:
        logger.warning("Services with old images: {}".format(", ".join(services_with_old_images)))



"""Get machine info data.

Uses PSUtil to get machine stats:
* CPU Percent
* Free Memory
* Memory Used
"""
import psutil
from buzio import formatStr
from dashing import dashing
from dashing.dashing import Text, VSplit

from megalus import __version__
from megalus.main import Megalus


def megalus_info_widget(meg: Megalus) -> VSplit:
    """Return Megalus Info Widget.

    :return: dashing VSplit instance
    """
    machine_info_widget = get_machine_info_widget()

    projects_in_error = [
        project
        for project in meg.config_data['compose_projects'].keys()
        if project not in meg.all_composes
    ]
    status_text = "All projects loaded." if not projects_in_error else formatStr.error(
        "Projects in error: {}".format(",".join(projects_in_error)), use_prefix=False)
    megalus_info_widget = Text(status_text, color=6, border_color=5, background_color=16,
                               title="Megalus v{}".format(__version__))
    return VSplit(megalus_info_widget, machine_info_widget)


def get_machine_info_widget() -> VSplit:
    """Return widget from machine stats.

    :return: dashing.HSplit
    """
    cpu_percent = round(psutil.cpu_percent(interval=None) * 10, 0) / 10
    free_memory = int(psutil.virtual_memory().available / 1024 / 1024)
    total_memory = int(psutil.virtual_memory().total / 1024 / 1024)
    memory_percent = (free_memory / total_memory) * 100
    free_space = round(psutil.disk_usage("/").free / 1024 / 1024 / 1024, 1)
    total_space = round(psutil.disk_usage(
        "/").total / 1024 / 1024 / 1024, 1)
    space_percent = (free_space / total_space) * 100

    if memory_percent > 100:
        memory_percent = 100

    if space_percent > 100:
        space_percent = 100

    if cpu_percent <= 50:
        cpu_color = 2
    elif cpu_percent <= 70:
        cpu_color = 3
    else:
        cpu_color = 1

    if memory_percent <= 20:
        memory_color = 1
    elif memory_percent <= 50:
        memory_color = 3
    else:
        memory_color = 2

    if space_percent <= 20:
        space_color = 1
    elif space_percent <= 50:
        space_color = 3
    else:
        space_color = 2

    return dashing.VSplit(
        dashing.HSplit(
            dashing.HGauge(
                val=cpu_percent,
                color=cpu_color,
                border_color=cpu_color,
                title="CPU:{}%".format(cpu_percent),
                background_color=16
            ),
            dashing.HGauge(
                val=memory_percent,
                color=memory_color,
                border_color=memory_color,
                title="Free Mem:{}M".format(free_memory),
                background_color=16
            ),
        ),
        dashing.HGauge(
            val=space_percent,
            color=space_color,
            border_color=space_color,
            title="Free Space:{}Gb".format(free_space),
            background_color=16
        )
    )



"""Test main code."""
from megalus.main import Megalus


def test_open_with_invalid_file(caplog):
    """Test open program with invalid file."""
    instance = Megalus(config_file="/path/not/found", logfile='/tmp/logfile')
    instance.get_services()
    running_command = [
        message
        for message in caplog.messages
        if "not found" in message
    ][0]
    assert "/path/not/found not found. Skipping..." in running_command



"""Utils module."""

from megalus.utils import get_ngrok_url


def test_get_ngrok_url(mocker, ngrok_response):
    """Test Get Ngrok URL."""
    ngrok_dict = {
        "port": 6544,
        "secure": False,
        "env": "MEGALUS_NGROK_TEST_ENV"
    }
    mocker.patch('megalus.utils.requests.get', return_value=ngrok_response)
    data = get_ngrok_url(ngrok_dict)

    expect_url = "http://87f3557f.ngrok.io"
    assert data == expect_url



"""Click command module."""
import os
import sys
from pathlib import Path

import click
import arrow
from loguru import logger

from megalus.bash.commands import bash
from megalus.buildcmd.commands import build
from megalus.check.commands import check
from megalus.commands.commands import config, install
from megalus.compose.commands import restart, scale, up
from megalus.down.commands import down
from megalus.logscmd.commands import logs
from megalus.main import Megalus
from megalus.run.commands import run
from megalus.status.commands import status
from megalus.stop.commands import stop
from megalus.start.commands import start


@click.group()
@click.option('--config_file', envvar='MEGALUS_PROJECT_CONFIG_FILE', required=True, type=click.Path())
@click.pass_context
def cli(ctx, config_file) -> None:
    """Define base click client.

    :param ctx: object: click context
    :param config_file: string: --config-file option
    :return: None
    """
    BASE_LOG_PATH = os.path.join(str(Path.home()), '.megalus', 'logscmd')

    if not os.path.exists(BASE_LOG_PATH):
        os.makedirs(BASE_LOG_PATH)

    now = arrow.utcnow().to("local").isoformat()
    LOGFILE = os.path.join(BASE_LOG_PATH, '{}.log'.format(now))

    DEFAULT_LOGGER_MESSAGE = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> |" \
                             " <level>{level: <8}</level> - <level>{message}</level>"

    config = {
        "handlers": [
            {"sink": sys.stdout, "format": DEFAULT_LOGGER_MESSAGE},
            {"sink": LOGFILE, "retention": "7 days", "format": DEFAULT_LOGGER_MESSAGE}
        ],
    }
    logger.configure(**config)
    meg = Megalus(config_file=config_file, logfile=LOGFILE)
    meg.get_services()
    ctx.obj = meg


cli.add_command(bash)
cli.add_command(build)
cli.add_command(config)
cli.add_command(down)
cli.add_command(install)
cli.add_command(logs)
cli.add_command(restart)
cli.add_command(run)
cli.add_command(scale)
cli.add_command(stop)
cli.add_command(up)
cli.add_command(check)
cli.add_command(status)
cli.add_command(start)


def start() -> None:
    """Start command.

    :return: None
    """
    cli()



"""Megalus Start Command."""
import click

from megalus import get_path
from megalus.main import Megalus


@click.command()
@click.argument('projects', nargs=-1, required=True)
@click.option('-d', is_flag=True)
@click.pass_obj
def start(meg: Megalus, projects: str, d: bool) -> None:
    """Start selected projects.

    :param meg: Megalus instance
    :param projects: Megalus projects to be started up
    :param d: detached option
    :return: None
    """
    options = "-d" if d or len(projects) > 1 else ""
    for project_name in meg.config_data['compose_projects']:
        if project_name in projects:
            compose_files = meg.config_data['compose_projects'][project_name]['files']
            compose_path = meg.config_data['compose_projects'][project_name]['path']
            meg.run_command(
                "cd {working_dir} && docker-compose -f {files} up {options}".format(
                    working_dir=get_path(compose_path, meg.base_path),
                    files=" -f ".join(compose_files),
                    options=options
                )
            )



"""Docker-Compose commands module."""
from typing import List, Optional

import click

from megalus.main import Megalus


def run_compose_command(meg: Megalus, action: str, service_data: dict, environment: Optional[List[str]] = None,
                        options: Optional[List[str]] = None, command_args: str = "") -> None:
    """Run docker-compose command.

    :param command_args: Command arguments to send after service name in docker-compose command.
    :param environment: Optional list of environment variables
    :param meg: Megalus instance
    :param action: docker-compose command
    :param service_data: docker service parsed data
    :param options: docker-compose command options
    :return: None
    """
    meg.run_command(
        "cd {working_dir} && {environment}docker-compose {files} {action}{options}{services}{args}".format(
            working_dir=service_data['working_dir'],
            environment="-e {}".format(" -e ".join(environment)) if environment else "",
            files="-f {}".format(" -f ".join(service_data['compose_files'])),
            options=" --{} ".format(" --".join(options)) if options else " ",
            action=action,
            services=service_data.get('name', ""),
            args=" {}".format(command_args) if command_args else ""
        )
    )


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.pass_obj
def restart(meg: Megalus, services: List[str]) -> None:
    """Restart selected services.

    :param meg: Megalus instance.
    :param services: Docker services
    :return: None
    """
    for service in services:
        service_data = meg.find_service(service)
        run_compose_command(meg, "restart", service_data)


@click.command()
@click.argument('service', required=True)
@click.argument('number', required=True, default=1, type=click.INT)
@click.pass_obj
def scale(meg: Megalus, service: str, number: int) -> None:
    """Scale selected services.

    :param meg: Megalus instance
    :param service: docker service to be scaled
    :param number: number of replicas
    :return: None
    """
    service_data = meg.find_service(service)
    options = ["scale {}={}".format(service_data['name'], number)]
    run_compose_command(meg, "up -d", options=options, service_data=service_data)


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.option('-d', is_flag=True)
@click.pass_obj
def up(meg: Megalus, services: List[str], d: bool) -> None:
    """Start selected services.

    :param meg: Megalus instance
    :param services: Services to be started up
    :param d: detached option
    :return: None
    """
    command = "up -d" if d or len(services) > 1 else "up"
    for service in services:
        service_data = meg.find_service(service)
        run_compose_command(meg, command, service_data=service_data)



"""Docker-Compose down module."""
from typing import List

import click

from megalus import get_path
from megalus.main import Megalus


@click.command()
@click.option('--remove-all', is_flag=True)
@click.argument('projects', nargs=-1)
@click.pass_obj
def down(meg: Megalus, projects: List[str], remove_all: bool) -> None:
    """Down compose project.

    :param meg: Megalus instance
    :param projects: projects to down
    :param remove_all: remove image and containers too
    :return: None
    """
    options = "--rmi all -v --remove-orphans" if remove_all else ""

    for project_name in meg.config_data['compose_projects']:
        if (projects and project_name in projects) or not projects:
            compose_files = meg.config_data['compose_projects'][project_name]['files']
            compose_path = meg.config_data['compose_projects'][project_name]['path']
            meg.run_command(
                "cd {working_dir} && docker-compose -f {files} down {options}".format(
                    working_dir=get_path(compose_path, meg.base_path),
                    files=" -f ".join(compose_files),
                    options=options
                )
            )



"""meg bash command."""

import click
from buzio import console
from loguru import logger

from megalus.compose.commands import run_compose_command
from megalus.main import Megalus
from megalus.utils import client, find_containers


@click.command()
@click.argument('service', nargs=1, required=True)
@click.pass_obj
def bash(meg: Megalus, service: str) -> None:
    """'bash' command.

    Use this command to open a bash session in current running container,
    or start one for select service.

    :param meg: click context object
    :param service: docker-compose service name
    :return: None
    """
    service_data = meg.find_service(service)

    container_id = None
    eligible_containers = find_containers(service_data['name'])
    if not eligible_containers:
        logger.info("Running /bin/bash in service {}".format(service_data['name']))
        run_compose_command(meg, action="run", options=["rm", "service-ports"], service_data=service_data,
                            command_args="/bin/bash")
    elif len(eligible_containers) == 1:
        container_id = eligible_containers[0].short_id
    else:
        container_names = [c.name for c in eligible_containers]
        container = console.choose(container_names, 'Please select the container')
        if container:
            container_id = client.containers.get(container).short_id
    if container_id:
        logger.info(
            "Running /bin/bash in service {} in container {}".format(service_data['name'], container_id))
        meg.run_command(
            'cd {} && docker exec -ti {} /bin/bash'.format(service_data['working_dir'], container_id)
        )



from codecs import open
from os import path
from setuptools import setup, find_packages
from pipenv.project import Project
from pipenv.utils import convert_deps_to_pip
from megalus import __version__

here = path.abspath(path.dirname(__file__))

pfile = Project(chdir=False).parsed_pipfile
requirements = convert_deps_to_pip(pfile['packages'], r=False)
test_requirements = convert_deps_to_pip(pfile['dev-packages'], r=False)


# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='megalus',
    version=__version__,
    description='Command line helpers for docker and docker-compose',
    long_description=long_description,
    author='Chris Maillefaud',
    include_package_data=True,
    # Choose your license
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='aws deploy docker npm redis memcached bash',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'meg=megalus.cmd:start'
        ],
    },
)
